'
' Interfacing to a Sure 2416 24*16 Dot Matrix Display Board
' Written by Les Johnson for the Proton BASIC compiler on a PIC18F device
'
Declare Reminders = Off
Declare Warnings = Off

$ifndef HT_CS
    $define HT_CS PORTB.2               ' Connects to the Sure2416 CS1 pin (Active Low)
$endif
$ifndef HT_WR
    $define HT_WR PORTB.1               ' Connects to the Sure2416 WR pin  (Active Low)
$endif
$ifndef HT_RD
    $define HT_RD PORTB.4               ' Connects to the Sure2416 RD pin  (Active Low)
$endif
$ifndef HT_Data
    $define HT_Data PORTB.3             ' Connects to the Sure2416 DATA pin
$endif
'
' Constants
'
' Commands written to the chip consist of a 3 bit "ID", followed by
' either 9 bits of "Command code" or 7 bits of address, plus 4 bits of data.
'
    Symbol cHT_ID_CMD 4                 ' ID = 100 - Commands
    Symbol cHT_ID_RD  6                 ' ID = 110 - Read RAM
    Symbol cHT_ID_WR  5                 ' ID = 101 - Write RAM

    Symbol cHT_CMD_SYSDIS = $00         ' CMD = 0000-0000-x     Turn off oscillator
    Symbol cHT_CMD_SYSON  = $01         ' CMD = 0000-0001-x     Enable system oscillator
    Symbol cHT_CMD_LEDOFF = $02         ' CMD = 0000-0010-x     LED duty cycle gen off
    Symbol cHT_CMD_LEDON  = $03         ' CMD = 0000-0011-x     LEDs On
    Symbol cHT_CMD_BLOFF  = $08         ' CMD = 0000-1000-x     Blink On
    Symbol cHT_CMD_BLON   = $09         ' CMD = 0000-1001-x     Blink Off
    Symbol cHT_CMD_SLVMD  = $10         ' CMD = 0001-0 = $x-x   Slave Mode
    Symbol cHT_CMD_MSTMD  = $14         ' CMD = 0001-01xx-x     Master Mode
    Symbol cHT_CMD_RCCLK  = $18         ' CMD = 0001-1 = $x-x   Use on-chip clock
    Symbol cHT_CMD_EXTCLK = $1C         ' CMD = 0001-11xx-x     Use external clock
    Symbol cHT_CMD_COMS00 = $20         ' CMD = 0010-ABxx-x     commons options
    Symbol cHT_CMD_COMS01 = $24         ' CMD = 0010-ABxx-x     commons options
    Symbol cHT_CMD_COMS10 = $28         ' CMD = 0010-ABxx-x     commons options
    Symbol cHT_CMD_COMS11 = $2C         ' CMD = 0010-ABxx-x     commons options
    Symbol cHT_CMD_PWM    = $A0         ' CMD = 101x-PPPP-x     PWM duty cycle

    Symbol cHT_XMax = 23
    Symbol cHT_YMax = 15

    Symbol tPixelClear = 0                  ' Set the pixel colour to off
    Symbol tPixelSet = 1                    ' Set the pixel colour to on
    Symbol tPixelXor = 2                    ' Set the pixel colour to xor
'
' Compiler system variables. Re-used by our routines
'
    Dim GEN As Byte System
    Dim GENH As Byte System
    Dim GEN2 As Byte System
    Dim GEN2H As Byte System
    Dim PP0 As Byte System
    Dim PP0H As Byte System
    Dim PP1 As Byte System
    Dim PP1H As Byte System
    Dim PP2 As Byte System
    Dim PP2H As Byte System
    Dim PP3 As Byte System
    Dim PP3H As Byte System
    Dim PP7 As Byte System
    Dim PP7H As Byte System
    Dim PP8 As Byte System
    Dim PP8H As Byte System
    Dim HT_wCharOffset As PRODL.Word               ' Alias the SFR pair PRODL and PRODH as a word variable
    Dim HT_wFSR1 As FSR1L.Word                     ' Alias the SFR pair FSR1L and FSR1H as a word variable
'
' Routine Variables
'
    Dim HT_bSureOutValue As Byte System
    Dim HT_bSureAddress As Byte System
    Dim HT_bWrBits As PRODL
    Dim HT_bWrFbits As PRODH
    Dim tPixelColor As Byte System                 ' Used to hold the colour of the pixel for all routines except DisplayChar
    Dim HT_bSetPixelX As Byte System               ' Used by the Plot/Pixel subroutines to hold the X position of the pixel
    Dim HT_bSetPixelY As Byte System               ' Used by the Plot/Pixel subroutines to hold the Y position of the pixel
    Dim HT_bAddress As Byte System                 ' Used by the Plot/Pixel subroutines to calculate the pixel address
    Dim HT_bBitVal As Byte System                  ' Used by the Plot/Pixel subroutines to calculate the pixel position withn memory
'
' Declare variables used by DisplayChar and DisplayString
'
    Dim HT_bCharX As GEN
    Dim HT_bCharY As GENH
    Dim HT_bCol As PP0
    Dim HT_bRow As PP0H
'
' Declare variables used by HT_Line
'
    Dim HT_bDeltaX As GEN
    Dim HT_bDeltaY As GENH
    Dim HT_bStepX As PP0
    Dim HT_bStepY As PP0H
    Dim HT_bDeviation As PP3
    Dim HT_bYpos_S As PP1
    Dim HT_bXpos_S As PP1H
    Dim HT_bYpos_E As PP2
    Dim HT_bXpos_E As PP2H
'
' Declare variables (aliases) used by HT_Circle
'
    Dim HT_bXX As GEN
    Dim HT_bYY As GENH
    Dim HT_bTR As PP0
    Dim HT_bDD As PP0H
    Dim HT_bQuadA As PP2
    Dim HT_bQuadB As PP2H
    Dim HT_bQuadC As PP3
    Dim HT_bQuadD As GEN2
    Dim HT_bRadius As GEN2H
'
' Keep a copy of the display controller's RAM so that we can know which bits are set without having to read the device.
'
    Dim HT_bShadowRam[96] As Byte Heap          ' Copy of the display's RAM
    Dim HT_tPixelColSave As Byte                ' Holds a copy of tPixelColor when impementing DispayChar

'-------------------------------------------------------------------------------------
    GoTo HT_StartUp                          ' Jump over the subroutines

'-------------------------------------------------------------------------------------
' Shift out data to the Sure 2416 board
' Input     : HT_bWrBits holds the value to send
'           : HT_bWrFbits (WREG) holds the starting bit of the value
' Output    : None
' Notes     : None
'
$define HT_ShiftOut(pData, pFirstBit) '
    HT_bWrBits = pData '
    WREG = pFirstBit   '
    _HT_ShiftOut

_HT_ShiftOut Macro
    GoSub __HT_ShiftOut
Endm

#ifMacro- _HT_ShiftOut, _HT_SendCommand, _HT_SendData, _HT_Init
__HT_ShiftOut:
    HT_bWrFbits = WREG
    While HT_bWrFbits <> 0
        Clear HT_WR
        WREG = HT_bWrBits & HT_bWrFbits
        If WREG <> 0 Then
            Set HT_Data
        Else
            Clear HT_Data
        EndIf
        Set HT_WR
        HT_bWrFbits = HT_bWrFbits >> 1
    Wend
    Return
#endIfMacro-

'-------------------------------------------------------------------------------------
' Send a command to the HT
' Input     : HT_bSureOutValue holds the command to send
' Output    : None
' Notes     : None
'
$define HT_SendCommand(pCommand) '
    WREG = pCommand '
    _HT_SendCommand

_HT_SendCommand Macro
    GoSub __HT_SendCommand
Endm

#ifMacro- _HT_SendCommand, _HT_Init
__HT_SendCommand:
    HT_bSureOutValue = WREG
    Clear HT_CS                                 ' Select chip
    HT_ShiftOut(cHT_ID_CMD, 1 << 2)             ' Send 3 bits of ID: Commmand
    HT_ShiftOut(HT_bSureOutValue, 1 << 7)       ' Send the actual command
    HT_ShiftOut(0, 1)                           ' One extra don't care bit in commands
    Set HT_CS                                   ' De-Select Chip
    Return
#endIfMacro-

'-------------------------------------------------------------------------------------
' Send a data byte to the HT
' Input     : HT_bSureOutValue holds the data to send
'             HT_bSureAddress holds the address
' Output    : None
' Notes     : None
'
$define HT_SendData(pAddress,pData) '
    HT_bSureAddress = pAddress '
    WREG = pData '
    _HT_SendData

_HT_SendData Macro
    GoSub __HT_SendData
Endm

#ifMacro- _HT_SendData, _HT_Init
__HT_SendData:
    HT_bSureOutValue = WREG                     ' Transfer WREG into it's working variable
    Clear HT_CS                                 ' Select chip
    HT_ShiftOut(cHT_ID_WR, 1 << 2)              ' Send ID: Write to RAM
    HT_ShiftOut(HT_bSureAddress, 1 << 6)        ' Send address
    HT_ShiftOut(HT_bSureOutValue, 1 << 3)       ' send 4 bits of data
    Set HT_CS                                   ' De-Select Chip
    Return
#endIfMacro-

'-------------------------------------------------------------------------------------
' Set or clear a pixel
' Input     : HT_bSetPixelX holds the X position of the pixel
'             HT_bSetPixelY holds the Y position of the Pixel
' Output    : None
' Notes     : Plot a point on the display, with the upper left hand corner being (0,0), and the lower right hand corner being (23, 15).
'             Y increases going "downward" in common with many displays
'
$define HT_Plot(pX, pY) '
    HT_bSetPixelX = pX '
    WREG = pY '
    _HT_Plot

_HT_Plot Macro
    GoSub __HT_SetPixel
Endm

#ifMacro- _HT_Plot, _HT_DisplayChar, _HT_Line, _HT_Circle
__HT_SetPixel:
    HT_bSetPixelY = WREG                            ' Transfer WREG into it's working variable
    If WREG.7 = 1 Then Return                       ' \
    If HT_bSetPixelY > cHT_YMax Then Return         ' |
    If HT_bSetPixelX.7 = 1 Then Return              ' | Boundary checks
    If HT_bSetPixelX > cHT_XMax Then Return         ' /
'
' The 4 bits in a single memory word go down, with the LSB (last transmitted) bit being on top.
' However, HT_ShiftOut() sends the LSB first, so we have to do a bit-reversal somewhere.
' This is done by shifting the single bit in the opposite direction from what you might expect.
'
    'HT_bBitVal = HT_bSetPixelY & 3                  ' \ Calculate which bit to act upon
    'HT_bBitVal = 8 >> HT_bBitVal                    ' /
    'HT_bAddress = HT_bSetPixelY / 4                 ' \
    'PRODL = HT_bSetPixelX * 4                       ' | Calculate which memory word this is in
    'HT_bAddress = HT_bAddress + PRODL               ' /
    HT_bBitVal = 8 >> (HT_bSetPixelY & 3)
    HT_bAddress=(HT_bSetPixelX << 2) + (HT_bSetPixelY >> 2)
    PRODL = HT_bShadowRam[HT_bAddress]               ' Speed things up by fetching from shadow RAM
    If tPixelColor = tPixelSet Then                 ' Do we need to set a pixel?
        PRODH = PRODL | HT_bBitVal                  ' Yes. So set the pixel
    ElseIf tPixelColor = 2 Then                     ' Do we need to Xor the pixel?
        PRODH = PRODL ^ HT_bBitVal                  ' Yes. So Xor the pixel
    Else                                            ' Otherwise...
        WREG = ~HT_bBitVal                          ' \
        PRODH = PRODL & WREG                        ' / Clear the pixel
    EndIf
    HT_bShadowRam[HT_bAddress] = PRODH               ' Place the new value into the shadow RAM
    HT_SendData(HT_bAddress, PRODH)                 ' Send the new memory value to the display
    Return
#endIfMacro-

'-------------------------------------------------------------------------------------
' Read a pixel's state
' Input     : HT_bSetPixelX holds the X position of the pixel
'             HT_bSetPixelY holds the Y position of the Pixel
' Output    : WREG holds 1 or 0 depending on whether the pixel is set or clear
'             Returns with WREG holding 0 if out of bounds
' Notes     : Looks up the value of a pixel from the shadow RAM for speed
'
$define HT_Pixel(pX, pY, pReturn) '
    HT_bSetPixelX = pX '
    WREG = pY '
    _HT_Pixel '
    pReturn = WREG

_HT_Pixel Macro
    GoSub __HT_GetPixel
Endm

#ifMacro- _HT_Pixel
__HT_GetPixel:
    HT_bSetPixelY = WREG                               ' Transfer WREG into it's working variable
    WREG = 0                                           ' Clear return ready for a boundary overflow
    If HT_bSetPixelY.7 = 1 Then Return                 ' \
    If HT_bSetPixelY > cHT_YMax Then Return            ' | Boundary checks
    If HT_bSetPixelX.7 = 1 Then Return                 ' | Return holding 0 if out of bounds
    If HT_bSetPixelX > cHT_XMax Then Return            ' /

    HT_bBitVal = HT_bSetPixelY & 3                     ' \ Calculate which bit to act upon
    HT_bBitVal = 8 >> HT_bBitVal                       ' /
    'HT_bAddress = HT_bSetPixelY / 4                    ' \
    'PRODL = HT_bSetPixelX * 4                          ' | Calculate which memory word this is in
    'HT_bAddress = HT_bAddress + PRODL                  ' /
    HT_bAddress=(HT_bSetPixelX << 2)+(HT_bSetPixelY >> 2)'computewhichmemorywordthisisin
    PRODL = HT_bShadowRam[HT_bAddress]                 ' Speed things up by fetching from shadow RAM
    WREG = PRODL & HT_bBitVal
    If WREG > 0 Then WREG = 1
    Return
#endIfMacro-

'-------------------------------------------------------------------------------------
' Draw or Erase a line from HT_bXpos_S, HT_bYpos_S to HT_bXpos_E, HT_bYpos_E
' Input     : HT_bXpos_S = Starting X position of the Line
'           : HT_bYpos_S = Starting Y position of the line
'           : HT_bXpos_E = Ending X position of the line
'           : HT_bYpos_E = Endif Y position of the line
'           : tPixelColor = Colour of the pixel
' Output    : None
' Notes     :
'
$define HT_Line(pXS,pYS,pXE,pYE) '
    HT_bXpos_S = pXS '
    HT_bYpos_S = pYS '
    HT_bXpos_E = pXE '
    WREG = pYE '
    _HT_Line

_HT_Line Macro
    GoSub __HT_DrawLine
Endm

#ifMacro- _HT_Line
__HT_DrawLine:
    HT_bYpos_E = WREG                               ' Transfer WREG into it's working variable
    HT_bDeltaY = HT_bYpos_E - HT_bYpos_S            ' Find the initial Y direction
    HT_bDeltaX = HT_bXpos_E - HT_bXpos_S            ' Find the initial X direction

    HT_bStepY = 1                                   ' Default to a positive step for Ypos
    If HT_bDeltaY.7 = 1 Then                        ' Check if negative
        HT_bDeltaY = -HT_bDeltaY                 ' Negate HT_bDeltaY
        HT_bStepY = -1                           ' And indicate a negative step
    EndIf
    HT_bStepX = 1                                   ' Default to a positive step for Xpos
    If HT_bDeltaX.7 = 1 Then                        ' Check if negative
        HT_bDeltaX = -HT_bDeltaX                 ' Negate HT_bDeltaX
        HT_bStepX = -1                           ' And indicate a negative step
    EndIf
    HT_Plot(HT_bXpos_S,HT_bYpos_S)                 ' The first pixel plot is a special case
    If HT_bDeltaX > HT_bDeltaY Then
        HT_bDeviation = HT_bDeltaY - HT_bDeltaX
        While HT_bXpos_S <> HT_bXpos_E
            If HT_bDeviation.7 = 0 Then             ' Check if negative
                HT_bYpos_S = HT_bYpos_S + HT_bStepY ' Move on the Y axis directed by HT_bStepY
                HT_bDeviation = HT_bDeviation - HT_bDeltaX
            EndIf
            HT_bXpos_S = HT_bXpos_S + HT_bStepX    ' Move on the X axis directed by HT_bStepX
            HT_bDeviation = HT_bDeviation + HT_bDeltaY
            HT_Plot(HT_bXpos_S,HT_bYpos_S)         ' Draw or Erase the pixel at this position
        Wend
        Return                                      ' Return prematurely (this saves an else command)
    EndIf
    HT_bDeviation = HT_bDeltaX - HT_bDeltaY

    While HT_bYpos_S <> HT_bYpos_E
        If HT_bDeviation.7 = 0 Then                 ' Check if negative
            HT_bXpos_S = HT_bXpos_S + HT_bStepX     ' Move on the X axis directed by HT_bStepX
            HT_bDeviation = HT_bDeviation - HT_bDeltaY
        EndIf
        HT_bYpos_S = HT_bYpos_S + HT_bStepY         ' Move on the Y axis directed by HT_bStepY
        HT_bDeviation = HT_bDeviation + HT_bDeltaX
        HT_Plot(HT_bXpos_S,HT_bYpos_S)             ' Draw or Erase the pixel at this position
    Wend
    Return
#endIfMacro-

'-------------------------------------------------------------------------------------
' Draw a circle
$define HT_Circle(pX, pY, pR) '
    HT_bXpos_S = pX '
    HT_bYpos_S = pY '
    WREG = pR '
    _HT_Circle

_HT_Circle Macro
    GoSub __HT_Circle
Endm

#ifMacro- _HT_Circle
__HT_Circle:
    HT_bRadius = WREG                              ' Transfer WREG into its working variable
    HT_bDD = HT_bYpos_S - HT_bXpos_S
    HT_bXX = 0
    HT_bYY = HT_bRadius
    HT_bTR = 3 - (2 * HT_bRadius)
    While HT_bXX <= HT_bYY
        HT_bQuadA = HT_bXpos_S + HT_bXX
        HT_bQuadB = HT_bYpos_S + HT_bYY

        HT_bQuadC = HT_bXpos_S - HT_bXX
        HT_bQuadD = HT_bYpos_S - HT_bYY
        GoSub HT_PlotIt
        HT_bQuadA = HT_bYpos_S + HT_bYY - HT_bDD
        HT_bQuadB = HT_bYpos_S + HT_bXX
        HT_bQuadC = HT_bYpos_S - HT_bYY - HT_bDD
        HT_bQuadD = HT_bYpos_S - HT_bXX
        GoSub HT_PlotIt
        If HT_bTR.7 = 1 Then
            HT_bTR = HT_bTR + 6
            HT_bTR = HT_bTR + (4 * HT_bXX)
        Else
            HT_bTR = HT_bTR + 10
            HT_bTR = HT_bTR + (4 * (HT_bXX - HT_bYY))
            Dec HT_bYY
        EndIf
        Inc HT_bXX
    Wend
    Return
'
' Plot 4 locations simultaneously
' This is part of the Circle subroutine
'
HT_PlotIt:
    HT_Plot(HT_bQuadA,HT_bQuadB)
    HT_Plot(HT_bQuadA,HT_bQuadD)
    HT_Plot(HT_bQuadC,HT_bQuadB)
    HT_Plot(HT_bQuadC,HT_bQuadD)
    Return
#endIfMacro-

'-------------------------------------------------------------------------------------
'Copyacharacterglyphfromthefontdatatabletodisplaymemory,withitsupperleftatthegivencoordinate
' Input     : HT_bCharX holds the X position of the character
'             HT_bCharY holds the Y position of the character
'             WREG holds the character to display
' Output    :WREG still holds the character sent
'Notes     : UsesHT_Plottodraweachdot of the character
'
$define HT_DisplayChar(pX,pY,pCh) '
    HT_bCharX = pX '
    HT_bCharY = pY '
    WREG = pCh '
    _HT_DisplayChar

_HT_DisplayChar Macro
    GoSub __HT_DisplayChar
Endm

#ifMacro- _HT_DisplayChar, _HT_DisplayString
__HT_DisplayChar:
    PP3H = WREG                                 ' Transfer WREG into its working variable
    HT_tPixelColSave = tPixelColor              ' Save the value of tPixelColor
'
'Movecharacterdefinition,pixelbypixel,ontothedisplay
' Fontsaredefinedasonebytepercolumn
'
' Characters's Address =  PP3H * 6
' PP3H holds the byte to display
' Add the offset for the address of font: wCharOffset + font label's address
' This is the code that requires altering for different sized fonts
'
    HT_bCol = 0
    Repeat                                      ' Form a loop for all the bytes in the character's font
        HT_wCharOffset = PP3H * 6               ' 6 bytes per font
        HT_wCharOffset = HT_wCharOffset + HT_bCol ' Add the offset to the character's value
        TABLAT = CRead8 Font_Table[HT_wCharOffset]    ' Read the byte from the font table
        '
        ' Send the value in TABLAT to the Display
        '
        HT_bRow = 7
        Repeat
            tPixelColor = TABLAT.6
            HT_Plot(HT_bCharX + HT_bCol,HT_bCharY + HT_bRow)
            TABLAT = TABLAT << 1
            Dec HT_bRow
        Until HT_bRow = 0
        Inc HT_bCol                             ' Next byte in the font
    Until HT_bCol >= 5                          ' Have we done 5 writes yet ?
    WREG = PP3H                                 ' Restore the value of WREG
    tPixelColor = HT_tPixelColSave               ' Restore the value of tPixelColor
    Return
#endIfMacro-

'-------------------------------------------------------------------------------------
' Display a string of characters on the LED matrix
'
' Input     : HT_bCharX holds the X position of the character
'             HT_bCharY holds the Y position of the character
'             SFR pair FSR1L\H hold the address of the string to display
' Output    : None
' Notes     : None
'
$define HT_DisplayString(pX,pY,pStr) '
    HT_bCharX = pX '
    WREG = pY '
    HT_wFSR1 = VarPtr pStr '
    _HT_DisplayString

_HT_DisplayString Macro
    GoSub __HT_DisplayString
Endm

#ifMacro- _HT_DisplayString
__HT_DisplayString:
    HT_bCharY = WREG                                ' Transfer WREG into its working variable
    HT_bStepX = 0                                   ' Reset a counter
    Repeat                                          ' Form a loop
        WREG = POSTINC1                             ' Transfer the element of the string into WREG and increment position
        Bz _HT_DisplayString_Exit                   ' Exit if a null found
        HT_DisplayChar(HT_bCharX, HT_bCharY, WREG)  ' Display the character held in WREG
        HT_bCharX = HT_bCharX + 6                   ' Move 6 pixels to the right (with of each character)
        Inc HT_bStepX                               ' Increment the counter
    Until HT_bStepX = 0                             ' Until 256 iterations have occured
_HT_DisplayString_Exit:
    Return
#endIfMacro-

'-------------------------------------------------------------------------------------
' Clear the display
'
' Input     : None
' Output    : None
' Notes     : None
'
$define HT_Clear() _HT_Clear
_HT_Clear Macro
    GoSub __HT_Clear
Endm

#ifMacro- _HT_Clear
__HT_Clear:
    Clear HT_bShadowRam                          ' Clear the shadow RAM
    HT_bDeltaX = 0
    Repeat                                      ' Create a loop
        HT_SendData(HT_bDeltaX,0)               ' Clear all the displays!
        Inc HT_bDeltaX
    Until HT_bDeltaX >= 127                     ' Until all are done
    Return
#endIfMacro-

'-------------------------------------------------------------------------------------
' Initialise the Sure 2416 display board
'
' Input     : None
' Output    : None
' Notes     : None
'
$define HT_Init() _HT_Init

_HT_Init Macro
    GoSub __HT_Init
Endm

#ifMacro- _HT_Init
__HT_Init:
    High HT_CS                                  ' Disable the board's interface
    High HT_WR                                  ' \ Place the board in write mode to start with
    Output HT_RD                                ' /
    Output HT_Data
    HT_SendCommand(cHT_CMD_SYSDIS)              ' Disable system
    HT_SendCommand(cHT_CMD_COMS11)              ' 16*32, PMOS drivers
    HT_SendCommand(cHT_CMD_MSTMD)               ' Master Mode
    HT_SendCommand(cHT_CMD_SYSON)               ' System on
    HT_SendCommand(cHT_CMD_LEDON)               ' LEDs on
    HT_Clear()                                  ' Clear the display
    tPixelColor = tPixelSet                     ' Default to the pixel being set
    Return
#endIfMacro-

'-------------------------------------------------------------------------------------
' Load the font for character display
#ifMacro- _HT_DisplayChar, _HT_DisplayString
    Include "Font.Inc"
#endIfMacro-

'-------------------------------------------------------------------------------------
HT_StartUp:
    HT_Init()                                   ' Initialise the display




